import re
from enum import Enum
from importlib import import_module
from typing import Any, Dict, List, Optional

from pytest import fixture, mark, param, raises

from omegaconf import (
    DictConfig,
    ListConfig,
    MissingMandatoryValue,
    OmegaConf,
    SCMode,
    open_dict,
)
from omegaconf._utils import _ensure_container
from omegaconf.errors import (
    InterpolationKeyError,
    InterpolationResolutionError,
    InterpolationToMissingValueError,
)
from tests import (
    B,
    Color,
    NestedInterpolationToMissing,
    StructuredInterpolationKeyError,
    User,
    Users,
    warns_dict_subclass_deprecated,
)


@mark.parametrize(
    "input_",
    [
        param([1, 2, 3], id="list"),
        param([1, 2, {"a": 3}], id="dict_in_list"),
        param([1, 2, [10, 20]], id="list_in_list"),
        param({"b": {"b": 10}}, id="dict_in_dict"),
        param({"b": [False, 1, "2", 3.0, Color.RED]}, id="list_in_dict"),
        param({"b": DictConfig(content=None)}, id="none_dictconfig"),
        param({"b": ListConfig(content=None)}, id="none_listconfig"),
        param({"b": DictConfig(content="???")}, id="missing_dictconfig"),
        param({"b": ListConfig(content="???")}, id="missing_listconfig"),
    ],
)
def test_to_container_returns_primitives(input_: Any) -> None:
    def assert_container_with_primitives(item: Any) -> None:
        if isinstance(item, list):
            for v in item:
                assert_container_with_primitives(v)
        elif isinstance(item, dict):
            for _k, v in item.items():
                assert_container_with_primitives(v)
        else:
            assert isinstance(item, (int, float, str, bool, type(None), Enum))

    c = OmegaConf.create(input_)
    res = OmegaConf.to_container(c, resolve=True)
    assert_container_with_primitives(res)


@mark.parametrize(
    "structured_config_mode,src,expected,key,expected_value_type",
    [
        param(
            SCMode.DICT,
            {"user": User(age=7, name="Bond")},
            {"user": {"name": "Bond", "age": 7}},
            "user",
            dict,
            id="DICT-dict",
        ),
        param(
            SCMode.DICT,
            [1, User(age=7, name="Bond")],
            [1, {"name": "Bond", "age": 7}],
            1,
            dict,
            id="DICT-list",
        ),
        param(
            SCMode.DICT_CONFIG,
            {"user": User(age=7, name="Bond")},
            {"user": User(age=7, name="Bond")},
            "user",
            DictConfig,
            id="DICT_CONFIG-dict",
        ),
        param(
            SCMode.DICT_CONFIG,
            [1, User(age=7, name="Bond")],
            [1, User(age=7, name="Bond")],
            1,
            DictConfig,
            id="DICT_CONFIG-list",
        ),
        param(
            SCMode.INSTANTIATE,
            {"user": User(age=7, name="Bond")},
            {"user": User(age=7, name="Bond")},
            "user",
            User,
            id="INSTANTIATE-dict",
        ),
        param(
            SCMode.INSTANTIATE,
            [1, User(age=7, name="Bond")],
            [1, User(age=7, name="Bond")],
            1,
            User,
            id="INSTANTIATE-list",
        ),
        param(
            None,
            {"user": User(age=7, name="Bond")},
            {"user": {"name": "Bond", "age": 7}},
            "user",
            dict,
            id="default-dict",
        ),
        param(
            None,
            [1, User(age=7, name="Bond")],
            [1, {"name": "Bond", "age": 7}],
            1,
            dict,
            id="default-list",
        ),
    ],
)
def test_scmode(
    src: Any,
    structured_config_mode: Optional[SCMode],
    expected: Any,
    expected_value_type: Any,
    key: Any,
) -> None:
    cfg = OmegaConf.create(src)
    if structured_config_mode is None:
        ret = OmegaConf.to_container(cfg)
    else:
        ret = OmegaConf.to_container(cfg, structured_config_mode=structured_config_mode)
    assert ret == expected
    assert isinstance(ret[key], expected_value_type)


@mark.parametrize(
    "src, expected, expected_with_resolve",
    [
        param([], None, None, id="empty_list"),
        param([1, 2, 3], None, None, id="list"),
        param([None], None, None, id="list_with_none"),
        param([1, "${0}", 3], None, [1, 1, 3], id="list_with_inter"),
        param({}, None, None, id="empty_dict"),
        param({"foo": "bar"}, None, None, id="dict"),
        param(
            {"foo": "${bar}", "bar": "zonk"},
            None,
            {"foo": "zonk", "bar": "zonk"},
            id="dict_with_inter",
        ),
        param({"foo": None}, None, None, id="dict_with_none"),
        param({"foo": "???"}, None, None, id="dict_missing_value"),
        param({"foo": None}, None, None, id="dict_none_value"),
        # containers
        param(
            {"foo": DictConfig(is_optional=True, content=None)},
            {"foo": None},
            None,
            id="dict_none_dictconfig",
        ),
        param(
            {"foo": DictConfig(content="???")},
            {"foo": "???"},
            None,
            id="dict_missing_dictconfig",
        ),
        param(
            {"foo": DictConfig(content="${bar}"), "bar": 10},
            {"foo": "${bar}", "bar": 10},
            {"foo": 10, "bar": 10},
            id="dict_inter_dictconfig",
        ),
        param(
            OmegaConf.create(
                {"foo": DictConfig(content="${bar}"), "bar": {"a": 0}}
            )._get_node("foo"),
            "${bar}",
            {"a": 0},
            id="toplevel_dict_inter",
        ),
        param(
            {"foo": ListConfig(content="???")},
            {"foo": "???"},
            None,
            id="dict_missing_listconfig",
        ),
        param(
            {"foo": ListConfig(is_optional=True, content=None)},
            {"foo": None},
            None,
            id="dict_none_listconfig",
        ),
        param(
            {"foo": ListConfig(content="${bar}"), "bar": 10},
            {"foo": "${bar}", "bar": 10},
            {"foo": 10, "bar": 10},
            id="dict_inter_listconfig",
        ),
        param(
            OmegaConf.create(
                {"foo": ListConfig(content="${bar}"), "bar": ["a"]}
            )._get_node("foo"),
            "${bar}",
            ["a"],
            id="toplevel_list_inter",
        ),
    ],
)
def test_to_container(src: Any, expected: Any, expected_with_resolve: Any) -> None:
    if expected is None:
        expected = src
    if expected_with_resolve is None:
        expected_with_resolve = expected
    cfg = _ensure_container(src)
    container = OmegaConf.to_container(cfg)
    assert container == expected
    container = OmegaConf.to_container(cfg, structured_config_mode=SCMode.INSTANTIATE)
    assert container == expected
    container = OmegaConf.to_container(cfg, resolve=True)
    assert container == expected_with_resolve


def test_to_container_invalid_input() -> None:
    with raises(
        ValueError,
        match=re.escape("Input cfg is not an OmegaConf config object (dict)"),
    ):
        OmegaConf.to_container({})


def test_string_interpolation_with_readonly_parent() -> None:
    cfg = OmegaConf.create({"a": 10, "b": {"c": "hello_${a}"}})
    OmegaConf.set_readonly(cfg, True)
    assert OmegaConf.to_container(cfg, resolve=True) == {
        "a": 10,
        "b": {"c": "hello_10"},
    }


class TestInstantiateStructuredConfigs:
    @fixture(
        params=[
            "tests.structured_conf.data.dataclasses",
            "tests.structured_conf.data.attr_classes",
        ]
    )
    def module(self, request: Any) -> Any:
        return import_module(request.param)

    def round_trip_to_object(self, input_data: Any) -> Any:
        serialized = OmegaConf.create(input_data)
        round_tripped = OmegaConf.to_object(serialized)
        return round_tripped

    def test_basic(self, module: Any) -> None:
        user = self.round_trip_to_object(module.User("Bond", 7))
        assert type(user) is module.User
        assert user.name == "Bond"
        assert user.age == 7

    def test_basic_with_missing(self, module: Any) -> None:
        with raises(MissingMandatoryValue):
            self.round_trip_to_object(module.User())

    def test_nested(self, module: Any) -> None:
        data = self.round_trip_to_object({"user": module.User("Bond", 7)})
        user = data["user"]
        assert type(user) is module.User
        assert user.name == "Bond"
        assert user.age == 7

    def test_nested_with_missing(self, module: Any) -> None:
        with raises(MissingMandatoryValue):
            self.round_trip_to_object({"user": module.User()})

    def test_list(self, module: Any) -> None:
        lst = self.round_trip_to_object(module.UserList([module.User("Bond", 7)]))
        assert type(lst) is module.UserList
        assert len(lst.list) == 1
        user = lst.list[0]
        assert type(user) is module.User
        assert user.name == "Bond"
        assert user.age == 7

    def test_list_with_missing(self, module: Any) -> None:
        with raises(MissingMandatoryValue):
            self.round_trip_to_object(module.UserList)

    def test_dict(self, module: Any) -> None:
        user_dict = self.round_trip_to_object(
            module.UserDict({"user007": module.User("Bond", 7)})
        )
        assert type(user_dict) is module.UserDict
        assert len(user_dict.dict) == 1
        user = user_dict.dict["user007"]
        assert type(user) is module.User
        assert user.name == "Bond"
        assert user.age == 7

    def test_dict_with_missing(self, module: Any) -> None:
        with raises(MissingMandatoryValue):
            self.round_trip_to_object(module.UserDict)

    def test_nested_object(self, module: Any) -> None:
        cfg = OmegaConf.structured(module.NestedConfig)
        # fill in missing values:
        cfg.default_value = module.NestedSubclass(mandatory_missing=123)
        cfg.user_provided_default.mandatory_missing = 456

        nested: Any = OmegaConf.to_object(cfg)
        assert type(nested) is module.NestedConfig
        assert type(nested.default_value) is module.NestedSubclass
        assert type(nested.user_provided_default) is module.Nested

        assert nested.default_value.mandatory_missing == 123
        assert nested.default_value.additional == 20
        assert nested.user_provided_default.mandatory_missing == 456

    def test_nested_object_with_missing(self, module: Any) -> None:
        with raises(MissingMandatoryValue):
            self.round_trip_to_object(module.NestedConfig)

    def test_to_object_resolve_is_True_by_default(self, module: Any) -> None:
        interp = self.round_trip_to_object(module.Interpolation)
        assert type(interp) is module.Interpolation

        assert interp.z1 == 100
        assert interp.z2 == "100_200"

    def test_to_container_INSTANTIATE_resolve_False(self, module: Any) -> None:
        """Test the lower level `to_container` API with SCMode.INSTANTIATE and resolve=False"""
        src = dict(
            obj=module.RelativeInterpolation(),
            interp_x="${obj.x}",
            interp_x_y="${obj.x}_${obj.x}",
        )
        nested = OmegaConf.create(src)
        container = OmegaConf.to_container(
            nested, resolve=False, structured_config_mode=SCMode.INSTANTIATE
        )
        assert isinstance(container, dict)
        assert container["interp_x"] == "${obj.x}"
        assert container["interp_x_y"] == "${obj.x}_${obj.x}"
        assert container["obj"].z1 == 100
        assert container["obj"].z2 == "100_200"

    def test_to_container_INSTANTIATE_enum_to_str_True(self, module: Any) -> None:
        """Test the lower level `to_container` API with SCMode.INSTANTIATE and resolve=False"""
        src = dict(
            color=Color.BLUE,
            obj=module.EnumOptional(),
        )
        nested = OmegaConf.create(src)
        container = OmegaConf.to_container(
            nested, enum_to_str=True, structured_config_mode=SCMode.INSTANTIATE
        )
        assert isinstance(container, dict)
        assert container["color"] == "BLUE"
        assert container["obj"].not_optional is Color.BLUE

    def test_to_object_InterpolationResolutionError(self, module: Any) -> None:
        with raises(InterpolationResolutionError):
            cfg = OmegaConf.structured(module.NestedWithAny)
            cfg.var.mandatory_missing = 123
            OmegaConf.to_object(cfg)

    def test_nested_object_with_Any_ref_type(self, module: Any) -> None:
        cfg = OmegaConf.structured(module.NestedWithAny())
        cfg.var.mandatory_missing = 123
        with open_dict(cfg):
            cfg.value_at_root = 456
        nested = self.round_trip_to_object(cfg)
        assert type(nested) is module.NestedWithAny

        assert type(nested.var) is module.Nested
        assert nested.var.with_default == 10
        assert nested.var.mandatory_missing == 123
        assert nested.var.interpolation == 456

    def test_str2user_instantiate(self, module: Any) -> None:
        with warns_dict_subclass_deprecated(module.DictSubclass.Str2User):
            cfg = OmegaConf.structured(module.DictSubclass.Str2User())
        cfg.bond = module.User(name="James Bond", age=7)
        data = self.round_trip_to_object(cfg)

        assert type(data) is module.DictSubclass.Str2User
        assert type(data.bond) is module.User
        assert data.bond == module.User("James Bond", 7)

    def test_str2user_with_field_instantiate(self, module: Any) -> None:
        with warns_dict_subclass_deprecated(module.DictSubclass.Str2UserWithField):
            cfg = OmegaConf.structured(module.DictSubclass.Str2UserWithField())
        cfg.mp = module.User(name="Moneypenny", age=11)
        data = self.round_trip_to_object(cfg)

        assert type(data) is module.DictSubclass.Str2UserWithField
        assert type(data.foo) is module.User
        assert data.foo == module.User("Bond", 7)
        assert type(data.mp) is module.User
        assert data.mp == module.User("Moneypenny", 11)

    def test_str2str_with_field_instantiate(self, module: Any) -> None:
        with warns_dict_subclass_deprecated(module.DictSubclass.Str2StrWithField):
            cfg = OmegaConf.structured(module.DictSubclass.Str2StrWithField())
        cfg.hello = "world"
        data = self.round_trip_to_object(cfg)

        assert type(data) is module.DictSubclass.Str2StrWithField
        assert data.foo == "bar"
        assert data.hello == "world"

    def test_setattr_for_user_with_extra_field(self, module: Any) -> None:
        cfg = OmegaConf.structured(module.User(name="James Bond", age=7))
        with open_dict(cfg):
            cfg.extra_field = 123

        user: Any = OmegaConf.to_object(cfg)
        assert type(user) is module.User
        assert user.extra_field == 123


class TestEnumToStr:
    """Test the `enum_to_str` argument to the `OmegaConf.to_container function`"""

    @mark.parametrize(
        "src,enum_to_str,expected",
        [
            param({Color.RED: "enum key"}, True, "RED", id="convert"),
            param({Color.RED: "enum key"}, False, Color.RED, id="dont-convert"),
        ],
    )
    def test_enum_to_str_for_keys(
        self, src: Any, enum_to_str: bool, expected: Any
    ) -> None:
        cfg = OmegaConf.create(src)
        container: Dict[Any, Any] = OmegaConf.to_container(cfg, enum_to_str=enum_to_str)  # type: ignore
        assert container == {expected: "enum key"}

    @mark.parametrize(
        "src,enum_to_str,expected",
        [
            param({"enum val": Color.RED}, True, "RED", id="convert"),
            param({"enum val": Color.RED}, False, Color.RED, id="dont-convert"),
        ],
    )
    def test_enum_to_str_for_values(
        self, src: Any, enum_to_str: bool, expected: Any
    ) -> None:
        cfg = OmegaConf.create(src)
        container: Dict[Any, Any] = OmegaConf.to_container(cfg, enum_to_str=enum_to_str)  # type: ignore
        assert container == {"enum val": expected}

    @mark.parametrize(
        "src,enum_to_str,expected",
        [
            param([Color.RED], True, "RED", id="convert"),
            param([Color.RED], False, Color.RED, id="dont-convert"),
        ],
    )
    def test_enum_to_str_for_list(
        self, src: Any, enum_to_str: bool, expected: Any
    ) -> None:
        cfg = OmegaConf.create(src)
        container: List[Any] = OmegaConf.to_container(cfg, enum_to_str=enum_to_str)  # type: ignore
        assert container == [expected]


@mark.parametrize(
    "cfg",
    [
        # to_container: throw_on_missing
        param(DictConfig("???"), id="dict:missing"),
        param(DictConfig({"a": "???"}), id="dict:missing_value"),
        param(DictConfig({"a": {"b": "???"}}), id="dict:nested"),
        param(ListConfig("???"), id="list:missing"),
        param(ListConfig(["???"]), id="list:missing_elt"),
        param(ListConfig(["abc", ["???"]]), id="list:nested"),
        param(OmegaConf.structured(B), id="structured:missing_field"),
        param(
            OmegaConf.structured(Users({"Bond": "???"})),  # type: ignore
            id="structured:missing_in_dict_field",
        ),
    ],
)
class TestThrowOnMissing:
    """Tests the `throw_on_missing` arugment to OmegaConf.to_container"""

    @mark.parametrize(
        "op",
        [
            param(
                lambda cfg: OmegaConf.to_container(cfg, throw_on_missing=True),
                id="to_container",
            ),
            param(OmegaConf.to_object, id="to_object"),
        ],
    )
    def test_throw_on_missing_raises(self, op: Any, cfg: Any) -> None:
        with raises(MissingMandatoryValue):
            op(cfg)

    def test_no_throw_on_missing(self, cfg: Any) -> None:
        assert OmegaConf.to_container(cfg, throw_on_missing=False) == cfg


@mark.parametrize(
    "cfg,expected,exception_type",
    [
        param(
            OmegaConf.create({"foo": "${bar}"}),
            {"foo": "${bar}"},
            InterpolationKeyError,
            id="interp_key_error",
        ),
        param(
            OmegaConf.create({"subcfg": {"x": "${missing}"}, "missing": "???"}).subcfg,
            {"x": "${missing}"},
            InterpolationToMissingValueError,
            id="interp_to_missing_in_dict",
        ),
        param(
            OmegaConf.create({"subcfg": ["${missing}"], "missing": "???"}).subcfg,
            ["${missing}"],
            InterpolationToMissingValueError,
            id="interp_to_missing_in_list",
        ),
        param(
            OmegaConf.structured(NestedInterpolationToMissing).subcfg,
            {"baz": "${..name}"},
            InterpolationToMissingValueError,
            id="interp_to_missing_in_structured",
        ),
        param(
            OmegaConf.structured(StructuredInterpolationKeyError),
            {"name": "${bar}"},
            InterpolationKeyError,
            id="interp_key_error_in_structured",
        ),
        param(
            DictConfig(content="${bar}"),
            "${bar}",
            InterpolationResolutionError,
            id="dictconfig_interp_key_error",
        ),
        param(
            ListConfig(content="${bar}"),
            "${bar}",
            InterpolationResolutionError,
            id="dictconfig_interp_key_error",
        ),
        param(
            OmegaConf.create({"foo": DictConfig(content="${bar}")}),
            {"foo": "${bar}"},
            InterpolationKeyError,
            id="dictconfig_interp_key_error_in_dict",
        ),
    ],
)
class TestResolveBadInterpolation:
    @mark.parametrize(
        "op",
        [
            param(
                lambda cfg: OmegaConf.to_container(
                    cfg, resolve=True, throw_on_missing=True
                ),
                id="throw_on_missing",
            ),
            param(
                lambda cfg: OmegaConf.to_container(
                    cfg, resolve=True, throw_on_missing=False
                ),
                id="no_throw_on_missing",
            ),
            param(OmegaConf.to_object, id="to_object"),
        ],
    )
    def test_resolve_bad_interpolation_raises(
        self, op: Any, cfg: Any, expected: Any, exception_type: Any
    ) -> None:
        with raises(exception_type):
            op(cfg)

    @mark.parametrize("throw_on_missing", [True, False])
    def test_resolve_bad_interpolation_no_resolve(
        self, cfg: Any, expected: Any, exception_type: Any, throw_on_missing: bool
    ) -> None:
        ret = OmegaConf.to_container(
            cfg, resolve=False, throw_on_missing=throw_on_missing
        )
        assert ret == expected



import os
import string
import warnings
from typing import Any, Optional

from omegaconf import Container, Node
from omegaconf._utils import _DEFAULT_MARKER_, _get_value
from omegaconf.basecontainer import BaseContainer
from omegaconf.errors import ConfigKeyError
from omegaconf.grammar_parser import parse
from omegaconf.resolvers.oc import dict


def create(obj: Any, _parent_: Container) -> Any:
    """Create a config object from `obj`, similar to `OmegaConf.create`"""
    from omegaconf import OmegaConf

    assert isinstance(_parent_, BaseContainer)
    return OmegaConf.create(obj, parent=_parent_)


def env(key: str, default: Any = _DEFAULT_MARKER_) -> Optional[str]:
    """
    :param key: Environment variable key
    :param default: Optional default value to use in case the key environment variable is not set.
                    If default is not a string, it is converted with str(default).
                    None default is returned as is.
    :return: The environment variable 'key'. If the environment variable is not set and a default is
            provided, the default is used. If used, the default is converted to a string with str(default).
            If the default is None, None is returned (without a string conversion).
    """
    try:
        return os.environ[key]
    except KeyError:
        if default is not _DEFAULT_MARKER_:
            return str(default) if default is not None else None
        else:
            raise KeyError(f"Environment variable '{key}' not found")


def decode(expr: Optional[str], _parent_: Container, _node_: Node) -> Any:
    """
    Parse and evaluate `expr` according to the `singleElement` rule of the grammar.

    If `expr` is `None`, then return `None`.
    """
    if expr is None:
        return None

    if not isinstance(expr, str):
        raise TypeError(
            f"`oc.decode` can only take strings or None as input, "
            f"but `{expr}` is of type {type(expr).__name__}"
        )

    parse_tree = parse(expr, parser_rule="singleElement", lexer_mode="VALUE_MODE")
    val = _parent_.resolve_parse_tree(parse_tree, node=_node_)
    return _get_value(val)


def deprecated(
    key: str,
    message: str = "'$OLD_KEY' is deprecated. Change your code and config to use '$NEW_KEY'",
    *,
    _parent_: Container,
    _node_: Node,
) -> Any:
    from omegaconf._impl import select_node

    if not isinstance(key, str):
        raise TypeError(
            f"oc.deprecated: interpolation key type is not a string ({type(key).__name__})"
        )

    if not isinstance(message, str):
        raise TypeError(
            f"oc.deprecated: interpolation message type is not a string ({type(message).__name__})"
        )

    full_key = _node_._get_full_key(key=None)
    target_node = select_node(_parent_, key, absolute_key=True)
    if target_node is None:
        raise ConfigKeyError(
            f"In oc.deprecated resolver at '{full_key}': Key not found: '{key}'"
        )
    new_key = target_node._get_full_key(key=None)
    msg = string.Template(message).safe_substitute(
        OLD_KEY=full_key,
        NEW_KEY=new_key,
    )
    warnings.warn(category=UserWarning, message=msg)
    return target_node


def select(
    key: str,
    default: Any = _DEFAULT_MARKER_,
    *,
    _parent_: Container,
) -> Any:
    from omegaconf._impl import select_value

    return select_value(cfg=_parent_, key=key, absolute_key=True, default=default)


__all__ = [
    "create",
    "decode",
    "deprecated",
    "dict",
    "env",
    "select",
]



import sys
from typing import Any, Dict, List, Optional, Tuple, Union

import attr
from pytest import importorskip

from omegaconf import II, MISSING, SI
from tests import Color

if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import TypedDict

# attr is a dependency of pytest which means it's always available when testing with pytest.
importorskip("attr")


class NotStructuredConfig:
    name: str = "Bond"
    age: int = 7

    def __eq__(self, other: Any) -> Any:
        if isinstance(other, type(self)):
            return self.name == other.name and self.age == other.age
        return False


if sys.version_info >= (3, 8):  # pragma: no cover

    class TypedDictSubclass(TypedDict):
        foo: int


@attr.s(auto_attribs=True)
class StructuredWithInvalidField:
    bar: NotStructuredConfig = NotStructuredConfig()


@attr.s(auto_attribs=True)
class User:
    name: str = MISSING
    age: int = MISSING


@attr.s(auto_attribs=True)
class UserList:
    list: List[User] = MISSING


@attr.s(auto_attribs=True)
class UserDict:
    dict: Dict[str, User] = MISSING


@attr.s(auto_attribs=True)
class UserWithDefaultName(User):
    name: str = "bob"


@attr.s(auto_attribs=True)
class MissingUserField:
    user: User = MISSING


@attr.s(auto_attribs=True)
class MissingUserWithDefaultNameField:
    user: UserWithDefaultName = MISSING


@attr.s(auto_attribs=True)
class OptionalUser:
    user: Optional[User] = None


@attr.s(auto_attribs=True)
class AnyTypeConfig:
    with_default: Any = "Can get any type at runtime"
    null_default: Any = None
    # Access to this prior to assigning a value to it will result in
    # a MissingMandatoryValue exception.
    # Equivalent to "???" in YAML files
    mandatory_missing: Any = MISSING

    # interpolation, will inherit the type and value of `with_default'
    interpolation: Any = II("with_default")

    # specific types assigned
    int_default: Any = 12
    float_default: Any = 10.0
    str_default: Any = "foobar"
    bool_default: Any = True
    enum_default: Any = Color.BLUE

    # test mixing with variable with a specific type annotation
    typed_int_default: int = 10


@attr.s(auto_attribs=True)
class BoolConfig:
    # with default value
    with_default: bool = True

    # default is None
    null_default: Optional[bool] = None

    # explicit no default
    mandatory_missing: bool = MISSING

    # interpolation, will inherit the type and value of `with_default'
    interpolation: bool = II("with_default")


@attr.s(auto_attribs=True)
class IntegersConfig:
    # with default value
    with_default: int = 10

    # default is None
    null_default: Optional[int] = None

    # explicit no default
    mandatory_missing: int = MISSING

    # interpolation, will inherit the type and value of `with_default'
    interpolation: int = II("with_default")


@attr.s(auto_attribs=True)
class StringConfig:
    # with default value
    with_default: str = "foo"

    # default is None
    null_default: Optional[str] = None

    # explicit no default
    mandatory_missing: str = MISSING

    # interpolation, will inherit the type and value of `with_default'
    interpolation: str = II("with_default")


@attr.s(auto_attribs=True)
class FloatConfig:
    # with default value
    with_default: float = 0.10

    # default is None
    null_default: Optional[float] = None

    # explicit no default
    mandatory_missing: float = MISSING

    # interpolation, will inherit the type and value of `with_default'
    interpolation: float = II("with_default")


@attr.s(auto_attribs=True)
class EnumConfig:
    # with default value
    with_default: Color = Color.BLUE

    # default is None
    null_default: Optional[Color] = None

    # explicit no default
    mandatory_missing: Color = MISSING

    # interpolation, will inherit the type and value of `with_default'
    interpolation: Color = II("with_default")


@attr.s(auto_attribs=True)
class ConfigWithList:
    list1: List[int] = [1, 2, 3]
    list2: Tuple[int, int, int] = (1, 2, 3)
    missing: List[int] = MISSING


@attr.s(auto_attribs=True)
class ConfigWithDict:
    dict1: Dict[str, Any] = {"foo": "bar"}
    missing: Dict[str, Any] = MISSING


@attr.s(auto_attribs=True)
class ConfigWithDict2:
    dict1: Dict[str, int] = {"foo": 2}


@attr.s(auto_attribs=True)
class Nested:
    # with default value
    with_default: int = 10

    # default is None
    null_default: Optional[int] = None

    # explicit no default
    mandatory_missing: int = MISSING

    # Note that since relative interpolations are not yet supported,
    # Nested configs and interpolations does not play too well together
    interpolation: int = II("value_at_root")


@attr.s(auto_attribs=True)
class NestedSubclass(Nested):
    additional: int = 20


@attr.s(auto_attribs=True)
class NestedConfig:
    default_value: Nested

    # with default value
    user_provided_default: Nested = Nested(with_default=42)

    value_at_root: int = 1000


@attr.s(auto_attribs=True)
class NestedWithAny:
    var: Any = Nested()


@attr.s(auto_attribs=True)
class NoDefaultValue:
    no_default: Any


@attr.s(auto_attribs=True)
class Interpolation:
    x: int = 100
    y: int = 200
    # The real type of y is int, cast the interpolation string
    # to help static type checkers to see this truth
    z1: int = II("x")
    z2: str = SI("${x}_${y}")


@attr.s(auto_attribs=True)
class RelativeInterpolation:
    x: int = 100
    y: int = 200
    z1: int = II(".x")
    z2: str = SI("${.x}_${.y}")


@attr.s(auto_attribs=True)
class BoolOptional:
    with_default: Optional[bool] = True
    as_none: Optional[bool] = None
    not_optional: bool = True


@attr.s(auto_attribs=True)
class IntegerOptional:
    with_default: Optional[int] = 1
    as_none: Optional[int] = None
    not_optional: int = 1


@attr.s(auto_attribs=True)
class FloatOptional:
    with_default: Optional[float] = 1.0
    as_none: Optional[float] = None
    not_optional: float = 1


@attr.s(auto_attribs=True)
class StringOptional:
    with_default: Optional[str] = "foo"
    as_none: Optional[str] = None
    not_optional: str = "foo"


@attr.s(auto_attribs=True)
class ListOptional:
    with_default: Optional[List[int]] = [1, 2, 3]
    as_none: Optional[List[int]] = None
    not_optional: List[int] = [1, 2, 3]


@attr.s(auto_attribs=True)
class TupleOptional:
    with_default: Optional[Tuple[int, int, int]] = (1, 2, 3)
    as_none: Optional[Tuple[int, int, int]] = None
    not_optional: Tuple[int, int, int] = (1, 2, 3)


@attr.s(auto_attribs=True)
class EnumOptional:
    with_default: Optional[Color] = Color.BLUE
    as_none: Optional[Color] = None
    not_optional: Color = Color.BLUE


@attr.s(auto_attribs=True)
class DictOptional:
    with_default: Optional[Dict[str, int]] = {"a": 10}
    as_none: Optional[Dict[str, int]] = None
    not_optional: Dict[str, int] = {"a": 10}


@attr.s(auto_attribs=True)
class RecursiveDict:
    d: Dict[str, "RecursiveDict"] = MISSING


@attr.s(auto_attribs=True)
class StructuredOptional:
    with_default: Optional[Nested] = Nested()
    as_none: Optional[Nested] = None
    not_optional: Nested = Nested()


@attr.s(auto_attribs=True, frozen=True)
class FrozenClass:
    user: User = User(name="Bart", age=10)
    x: int = 10
    list: List[int] = [1, 2, 3]


@attr.s(auto_attribs=True)
class ContainsFrozen:
    x: int = 10
    frozen: FrozenClass = FrozenClass()


@attr.s(auto_attribs=True)
class WithListField:
    list: List[int] = [1, 2, 3]


@attr.s(auto_attribs=True)
class WithDictField:
    dict: Dict[str, int] = {"foo": 10, "bar": 20}


if sys.version_info >= (3, 8):  # pragma: no cover

    @attr.s(auto_attribs=True)
    class WithTypedDictField:
        dict: TypedDictSubclass


@attr.s(auto_attribs=True)
class ErrorDictObjectKey:
    # invalid dict key, must be str
    dict: Dict[object, str] = {object(): "foo", object(): "bar"}


class RegularClass:
    pass


@attr.s(auto_attribs=True)
class ErrorDictUnsupportedValue:
    # invalid dict value type, not one of the supported types
    dict: Dict[str, RegularClass] = {}


@attr.s(auto_attribs=True)
class ErrorListUnsupportedValue:
    # invalid dict value type, not one of the supported types
    dict: List[RegularClass] = []


@attr.s(auto_attribs=True)
class ListExamples:
    any: List[Any] = [1, "foo"]
    ints: List[int] = [1, 2]
    strings: List[str] = ["foo", "bar"]
    booleans: List[bool] = [True, False]
    colors: List[Color] = [Color.RED, Color.GREEN]


@attr.s(auto_attribs=True)
class TupleExamples:
    any: Tuple[Any, Any] = (1, "foo")
    ints: Tuple[int, int] = (1, 2)
    strings: Tuple[str, str] = ("foo", "bar")
    booleans: Tuple[bool, bool] = (True, False)
    colors: Tuple[Color, Color] = (Color.RED, Color.GREEN)


@attr.s(auto_attribs=True)
class DictExamples:
    any: Dict[str, Any] = {"a": 1, "b": "foo"}
    ints: Dict[str, int] = {"a": 10, "b": 20}
    strings: Dict[str, str] = {"a": "foo", "b": "bar"}
    booleans: Dict[str, bool] = {"a": True, "b": False}
    colors: Dict[str, Color] = {
        "red": Color.RED,
        "green": Color.GREEN,
        "blue": Color.BLUE,
    }
    int_keys: Dict[int, str] = {1: "one", 2: "two"}
    float_keys: Dict[float, str] = {1.1: "one", 2.2: "two"}
    bool_keys: Dict[bool, str] = {True: "T", False: "F"}
    enum_key: Dict[Color, str] = {Color.RED: "red", Color.GREEN: "green"}


@attr.s(auto_attribs=True)
class DictOfObjects:
    users: Dict[str, User] = {"joe": User(name="Joe", age=18)}


@attr.s(auto_attribs=True)
class ListOfObjects:
    users: List[User] = [User(name="Joe", age=18)]


class DictSubclass:
    @attr.s(auto_attribs=True)
    class Str2Str(Dict[str, str]):
        pass

    @attr.s(auto_attribs=True)
    class Str2Int(Dict[str, int]):
        pass

    @attr.s(auto_attribs=True)
    class Int2Str(Dict[int, str]):
        pass

    @attr.s(auto_attribs=True)
    class Float2Str(Dict[float, str]):
        pass

    @attr.s(auto_attribs=True)
    class Bool2Str(Dict[bool, str]):
        pass

    @attr.s(auto_attribs=True)
    class Color2Str(Dict[Color, str]):
        pass

    @attr.s(auto_attribs=True)
    class Color2Color(Dict[Color, Color]):
        pass

    @attr.s(auto_attribs=True)
    class Str2User(Dict[str, User]):
        pass

    @attr.s(auto_attribs=True)
    class Str2StrWithField(Dict[str, str]):
        foo: str = "bar"

    @attr.s(auto_attribs=True)
    class Str2IntWithStrField(Dict[str, int]):
        foo: int = 1

    @attr.s(auto_attribs=True)
    class Str2UserWithField(Dict[str, User]):
        foo: User = User("Bond", 7)

    class Error:
        @attr.s(auto_attribs=True)
        class User2Str(Dict[User, str]):
            pass


@attr.s(auto_attribs=True)
class Plugin:
    name: str = MISSING
    params: Any = MISSING


@attr.s(auto_attribs=True)
class ConcretePlugin(Plugin):
    name: str = "foobar_plugin"

    @attr.s(auto_attribs=True)
    class FoobarParams:
        foo: int = 10

    params: FoobarParams = FoobarParams()


@attr.s(auto_attribs=True)
class PluginWithAdditionalField(Plugin):
    name: str = "foobar2_plugin"
    additional: int = 10


# Does not extend Plugin, cannot be assigned or merged
@attr.s(auto_attribs=True)
class FaultyPlugin:
    name: str = "faulty_plugin"


@attr.s(auto_attribs=True)
class PluginHolder:
    none: Optional[Plugin] = None
    missing: Plugin = MISSING
    plugin: Plugin = Plugin()
    plugin2: Plugin = ConcretePlugin()


@attr.s(auto_attribs=True)
class LinkedList:
    next: Optional["LinkedList"] = None
    value: Any = MISSING


@attr.s(auto_attribs=True)
class RecursiveList:
    d: List["RecursiveList"] = MISSING


class MissingTest:
    @attr.s(auto_attribs=True)
    class Missing1:
        head: LinkedList = MISSING

    @attr.s(auto_attribs=True)
    class Missing2:
        head: LinkedList = LinkedList(next=MISSING, value=1)


@attr.s(auto_attribs=True)
class NestedWithNone:
    plugin: Optional[Plugin] = None


@attr.s(auto_attribs=True)
class UnionError:
    x: Union[int, str] = 10


@attr.s(auto_attribs=True)
class WithNativeMISSING:
    num: int = attr.NOTHING  # type: ignore


@attr.s(auto_attribs=True)
class MissingStructuredConfigField:
    plugin: Plugin = MISSING


@attr.s(auto_attribs=True)
class ListClass:
    list: List[int] = []
    tuple: Tuple[int, int] = (1, 2)


@attr.s(auto_attribs=True)
class UntypedList:
    list: List = [1, 2]  # type: ignore
    opt_list: Optional[List] = None  # type: ignore


@attr.s(auto_attribs=True)
class UntypedDict:
    dict: Dict = {"foo": "var"}  # type: ignore
    opt_dict: Optional[Dict] = None  # type: ignore



# Generated from D:\bld\omegaconf_1636817949024\work\omegaconf\grammar\OmegaConfGrammarParser.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .OmegaConfGrammarParser import OmegaConfGrammarParser
else:
    from OmegaConfGrammarParser import OmegaConfGrammarParser

# This class defines a complete generic visitor for a parse tree produced by OmegaConfGrammarParser.

class OmegaConfGrammarParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by OmegaConfGrammarParser#configValue.
    def visitConfigValue(self, ctx:OmegaConfGrammarParser.ConfigValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#singleElement.
    def visitSingleElement(self, ctx:OmegaConfGrammarParser.SingleElementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#text.
    def visitText(self, ctx:OmegaConfGrammarParser.TextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#element.
    def visitElement(self, ctx:OmegaConfGrammarParser.ElementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#listContainer.
    def visitListContainer(self, ctx:OmegaConfGrammarParser.ListContainerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#dictContainer.
    def visitDictContainer(self, ctx:OmegaConfGrammarParser.DictContainerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#dictKeyValuePair.
    def visitDictKeyValuePair(self, ctx:OmegaConfGrammarParser.DictKeyValuePairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#sequence.
    def visitSequence(self, ctx:OmegaConfGrammarParser.SequenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#interpolation.
    def visitInterpolation(self, ctx:OmegaConfGrammarParser.InterpolationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#interpolationNode.
    def visitInterpolationNode(self, ctx:OmegaConfGrammarParser.InterpolationNodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#interpolationResolver.
    def visitInterpolationResolver(self, ctx:OmegaConfGrammarParser.InterpolationResolverContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#configKey.
    def visitConfigKey(self, ctx:OmegaConfGrammarParser.ConfigKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#resolverName.
    def visitResolverName(self, ctx:OmegaConfGrammarParser.ResolverNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#quotedValue.
    def visitQuotedValue(self, ctx:OmegaConfGrammarParser.QuotedValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#primitive.
    def visitPrimitive(self, ctx:OmegaConfGrammarParser.PrimitiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by OmegaConfGrammarParser#dictKey.
    def visitDictKey(self, ctx:OmegaConfGrammarParser.DictKeyContext):
        return self.visitChildren(ctx)



del OmegaConfGrammarParser


# Generated from D:\bld\omegaconf_1636817949024\work\omegaconf\grammar\OmegaConfGrammarParser.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .OmegaConfGrammarParser import OmegaConfGrammarParser
else:
    from OmegaConfGrammarParser import OmegaConfGrammarParser

# This class defines a complete listener for a parse tree produced by OmegaConfGrammarParser.
class OmegaConfGrammarParserListener(ParseTreeListener):

    # Enter a parse tree produced by OmegaConfGrammarParser#configValue.
    def enterConfigValue(self, ctx:OmegaConfGrammarParser.ConfigValueContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#configValue.
    def exitConfigValue(self, ctx:OmegaConfGrammarParser.ConfigValueContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#singleElement.
    def enterSingleElement(self, ctx:OmegaConfGrammarParser.SingleElementContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#singleElement.
    def exitSingleElement(self, ctx:OmegaConfGrammarParser.SingleElementContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#text.
    def enterText(self, ctx:OmegaConfGrammarParser.TextContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#text.
    def exitText(self, ctx:OmegaConfGrammarParser.TextContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#element.
    def enterElement(self, ctx:OmegaConfGrammarParser.ElementContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#element.
    def exitElement(self, ctx:OmegaConfGrammarParser.ElementContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#listContainer.
    def enterListContainer(self, ctx:OmegaConfGrammarParser.ListContainerContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#listContainer.
    def exitListContainer(self, ctx:OmegaConfGrammarParser.ListContainerContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#dictContainer.
    def enterDictContainer(self, ctx:OmegaConfGrammarParser.DictContainerContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#dictContainer.
    def exitDictContainer(self, ctx:OmegaConfGrammarParser.DictContainerContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#dictKeyValuePair.
    def enterDictKeyValuePair(self, ctx:OmegaConfGrammarParser.DictKeyValuePairContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#dictKeyValuePair.
    def exitDictKeyValuePair(self, ctx:OmegaConfGrammarParser.DictKeyValuePairContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#sequence.
    def enterSequence(self, ctx:OmegaConfGrammarParser.SequenceContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#sequence.
    def exitSequence(self, ctx:OmegaConfGrammarParser.SequenceContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#interpolation.
    def enterInterpolation(self, ctx:OmegaConfGrammarParser.InterpolationContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#interpolation.
    def exitInterpolation(self, ctx:OmegaConfGrammarParser.InterpolationContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#interpolationNode.
    def enterInterpolationNode(self, ctx:OmegaConfGrammarParser.InterpolationNodeContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#interpolationNode.
    def exitInterpolationNode(self, ctx:OmegaConfGrammarParser.InterpolationNodeContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#interpolationResolver.
    def enterInterpolationResolver(self, ctx:OmegaConfGrammarParser.InterpolationResolverContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#interpolationResolver.
    def exitInterpolationResolver(self, ctx:OmegaConfGrammarParser.InterpolationResolverContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#configKey.
    def enterConfigKey(self, ctx:OmegaConfGrammarParser.ConfigKeyContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#configKey.
    def exitConfigKey(self, ctx:OmegaConfGrammarParser.ConfigKeyContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#resolverName.
    def enterResolverName(self, ctx:OmegaConfGrammarParser.ResolverNameContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#resolverName.
    def exitResolverName(self, ctx:OmegaConfGrammarParser.ResolverNameContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#quotedValue.
    def enterQuotedValue(self, ctx:OmegaConfGrammarParser.QuotedValueContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#quotedValue.
    def exitQuotedValue(self, ctx:OmegaConfGrammarParser.QuotedValueContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#primitive.
    def enterPrimitive(self, ctx:OmegaConfGrammarParser.PrimitiveContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#primitive.
    def exitPrimitive(self, ctx:OmegaConfGrammarParser.PrimitiveContext):
        pass


    # Enter a parse tree produced by OmegaConfGrammarParser#dictKey.
    def enterDictKey(self, ctx:OmegaConfGrammarParser.DictKeyContext):
        pass

    # Exit a parse tree produced by OmegaConfGrammarParser#dictKey.
    def exitDictKey(self, ctx:OmegaConfGrammarParser.DictKeyContext):
        pass



del OmegaConfGrammarParser